import tempfile
import zipfile
from pathlib import Path
import uuid
from tqdm import tqdm
import requests


def download_large_file(url, local_file_path):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        
        total_size_in_bytes = int(r.headers.get('content-length', 0))
        progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
        
        with open(str(local_file_path), 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192): 
                progress_bar.update(len(chunk))
                f.write(chunk)
                
    return local_file_path


def download_model_from_s3(public_url, directory_path):
    directory_path = Path(directory_path)
    
    if not directory_path.exists():
        directory_path.mkdir(parents=True, exist_ok=True) 
    
    temp_path = Path(tempfile.gettempdir()) / "{}.zip".format(uuid.uuid4())
    download_large_file(public_url, temp_path)
    
    with zipfile.ZipFile(temp_path, 'r') as zip_ref:
        zip_ref.extractall(str(directory_path))



import click
from stochasticx.utils.docker import (
    start_container, 
    stop_and_remove_container,
    get_logs_container,
    get_open_ports_container
)
from stochasticx.utils.gpu_utils import is_nvidia_gpu_available, get_gpu_info, get_cpu_info
from stochasticx.local.stable_diffusion import (
    inference,
    init,
    generate_images
)
import sys
from PIL import Image
import numpy as np
from pathlib import Path
import uuid
from stochasticx.datasets.datasets import Datasets
from stochasticx.local.download_models import download_model_from_s3
from stochasticx.utils.logging import configure_logger
from stochasticx.deployment.deployments import (
    StableDiffusionDeployments,
    StableDiffusionDeployment
)


logger = configure_logger(__name__)

@click.group(name="stable-diffusion")
def stable_diffusion():
    pass


@click.command(name="download")
@click.option(
    '--type', 
    default="pytorch", 
    show_default=True, 
    help='Model type', 
    type=click.Choice([
        'pytorch', 
        'onnx',
        'tensorrt',
        'nvfuser',
        'flash_attention'
    ])
)
@click.option(
    '--local_dir_path', 
    default="downloaded_models", 
    show_default=True, 
    help='Path where the model will be downloaded'
)
def download(type, local_dir_path):
    click.secho("\n[+] Downloading model\n", fg='blue', bold=True) 
    download_model_from_s3(
        "https://stochasticai.s3.amazonaws.com/stable-diffusion/{}_model.zip".format(type),
        local_dir_path
    )
    click.secho("[+] Model downloaded\n", fg='blue', bold=True) 
    

@click.command(name="deploy")
@click.option(
    '--type', 
    default="aitemplate", 
    show_default=True, 
    help='Model type', 
    type=click.Choice([
        'pytorch', 
        'aitemplate', 
        'tensorrt',
        'nvfuser',
        'flash_attention',
        'onnx_cuda',
        'auto',
        'cloud'
    ], case_sensitive=False)
)
@click.option(
    '--port', 
    default="5000", 
    show_default=True, 
    help='Port'
)
def deploy(type, port):    
    try:
        Datasets.get_datasets()
    except:
        click.secho("\n[+] Execute ---> stochasticx login", fg='red', bold=True) 
        click.secho("[+] Or sign up in the following URL https://app.stochastic.ai/signup \n", fg='red', bold=True)
        sys.exit()
        
    if type == "cloud":
        click.secho("\n[+] Deploying the Stable-Diffusion model. It might take some minutes", fg='blue', bold=True)
        click.secho("[+] List your deployments with the following command: stochasticx stable-diffusion list\n", fg='blue', bold=True)
        StableDiffusionDeployments.deploy()
        
    else: # local
        gpu2type = {"a100":"aitemplate",
                    "t4":"tensorrt"}
        gpu_info = get_gpu_info()[0]
        
        # If there is no GPU
        if gpu_info is None or len(gpu_info) == 0:
            logger.warning("Please install gpu driver or use the machine which has gpu")
            return
        if gpu_info["name"] not in gpu2type:
            logger.warning("Currently we only support a100 and t4 gpus. We may support other gpus in the future")
            return
        
        # If there is GPU
        if type == "auto":
            type = gpu2type[gpu_info["name"]]
            click.secho("[+] Your gpu is {}, the best option is {}".format(gpu_info["name"],type), fg='blue', bold=True)
        if type == "tensorrt":        
            docker_image = "public.ecr.aws/t8g5g2q5/stable-diffusion:{}_{}".format(type,gpu_info["name"].lower())
        else:
            docker_image = "public.ecr.aws/t8g5g2q5/stable-diffusion:{}".format(type)

        click.secho("[+] Deploying Stable Diffusion model", fg='blue', bold=True)
        click.secho("[+] If it is the first time you deploy the model, it might take some minutes to deploy it", fg='blue', bold=True)
        start_container(
            docker_image=docker_image,
            ports={"5000": port},
            container_name="stochasticx_stable_diffusion",
            detach=True,
            gpu=is_nvidia_gpu_available()
        )
        
        click.secho("[+] Stable Diffusion running in the port {}".format(port), fg='blue', bold=True)
        click.secho("[+] Using GPU: {}".format(is_nvidia_gpu_available()), fg='blue', bold=True)
        click.secho("[+] Run the following command to start generating:", fg='blue', bold=True)
        click.secho("\tstochasticx stable-diffusion infer --prompt 'an astronaut riding a horse'", fg='blue', bold=True)
        
        
@click.command(name="ls")
@click.option(
    '--id', 
    help='The ID of the deployment'
)
def list(id):
    click.secho("\n[+] Deployments\n", fg='blue', bold=True)
    
    if id is not None:
        deployment = StableDiffusionDeployments.get_deployment(id)
        print(deployment)
    else:
        deployments = StableDiffusionDeployments.get_deployments()
        
        for deployment in deployments:
            print(deployment)


@click.command(name="logs")
def logs():
    click.secho("\n[+] Logs\n", fg='blue', bold=True)
    logs = get_logs_container("stochasticx_stable_diffusion")
    print(logs)


@click.command(name="stop")
@click.option(
    '--cloud_id', 
    help='The ID of the deployment from the cloud'
)
def stop(cloud_id):
    click.secho("[+] Stopping and removing stable-diffusion model", fg='blue', bold=True)
    if cloud_id is not None:
        StableDiffusionDeployments.delete(cloud_id)
    else:
        stop_and_remove_container("stochasticx_stable_diffusion")
    
    click.secho("[+] Removed", fg='green', bold=True)
    
    
@click.command(name="infer")
@click.option(
    '--prompt', 
    required=True, 
    help='Prompt to generate images', 
)
@click.option(
    '--cloud_id', 
    required=False, 
    help='Cloud ID deployment', 
)
@click.option(
    '--img_height', 
    default=512, 
    type=int,
    show_default=True, 
    help='The height in pixels of the generated image.'
)
@click.option(
    '--img_width', 
    default=512, 
    type=int,
    show_default=True, 
    help='The width in pixels of the generated image.'
)
@click.option(
    '--num_inference_steps', 
    default=50, 
    type=int,
    show_default=True, 
    help='The number of denoising steps. More denoising steps usually lead to a higher quality image at the expense of slower inference'
)
@click.option(
    '--num_images_per_prompt', 
    default=1, 
    type=int,
    show_default=True, 
    help='The number of images to generate per prompt.'
)
@click.option(
    '--seed', 
    default=None, 
    type=int,
    show_default=True, 
    help='Seed to make generation deterministic'
)
@click.option(
    '--saving_path', 
    default="generated_images", 
    type=str,
    show_default=True, 
    help='Directory where the generated images will be saved'
)
def infer(
    prompt, 
    cloud_id,
    img_height, 
    img_width, 
    num_inference_steps,
    num_images_per_prompt,
    seed,
    saving_path
):
    # Create directory to save images if it does not exist
    saving_path = Path(saving_path)
    if not saving_path.exists():
        saving_path.mkdir(exist_ok=True, parents=True)   
            
    if cloud_id is not None:
        deployment = StableDiffusionDeployments.get_deployment(cloud_id)
        url = deployment.client_url
        headers = {
            "apiKey": deployment.api_key
        }
    else:                        
        ports = get_open_ports_container("stochasticx_stable_diffusion")
        if len(ports) == 0:
            print("[+] Before doing the inference you have to run: stochasticx stable-diffusion deploy")
            sys.exit()
            
        url = "http://127.0.0.1:{}/predict".format(list(ports.values())[0])
        headers = {}
        
    click.secho("[+] Generating images...", fg='blue', bold=True)
    images, time = inference(
        url=url,
        headers=headers,
        prompt=prompt, 
        img_height=img_height, 
        img_width=img_width, 
        num_inference_steps=num_inference_steps,
        num_images_per_prompt=num_images_per_prompt,
        seed=seed
    )
    
    click.secho("[+] Time needed to generate the images: {} seconds".format(time), fg='blue', bold=True)
    
    pil_images = []
    for img in images:
        pil_images.append(
            Image.fromarray(np.uint8(img))
        )
            
    # Save PIL images with a random name
    for img in pil_images:
        img.save('{}/{}.png'.format(
            saving_path.as_posix(),
            uuid.uuid4()
        ))

    click.secho("[+] Images saved in the following path: {}".format(saving_path.as_posix()), fg='green', bold=True)


stable_diffusion.add_command(download)
stable_diffusion.add_command(deploy)
stable_diffusion.add_command(logs)
stable_diffusion.add_command(stop)
stable_diffusion.add_command(list)
stable_diffusion.add_command(infer)


import click
from stochasticx.auth.auth import Stochastic
from getpass import getpass

stochastic = Stochastic()

@click.command(name="login")
@click.option('--username', help='Username for login')
@click.option('--password', help='Password for login')
def login(username, password):
    if username is None:
        username = input("Enter your email: ")
    
    if password is None:
        password = getpass()
        
    stochastic.login(username, password)
        
    click.secho("[+] Login successfully", fg="green", bold=True)


@click.group(name="me")
def me():
    pass


@click.command(name="profile")
def profile():
    click.secho("\n[+] Collecting profile information", fg='blue', bold=True)
    profile_info = stochastic.get_profile()
    click.secho(profile_info, bold=True)
    

@click.command(name="company")
def company():
    company_info = stochastic.get_company()
    click.secho("\n[+] Collecting company information", fg='blue', bold=True)
    click.secho(company_info, bold=True)
    

@click.command(name="usage")
def usage():
    usage_info = stochastic.get_usage_quota()
    click.secho("\n[+] Collecting usage information", fg='blue', bold=True)
    click.secho(usage_info, bold=True)
    

me.add_command(profile)
me.add_command(company)
me.add_command(usage)


from pathlib import Path

import click
from stochasticx.datasets.datasets import Datasets, Dataset
from stochasticx.utils.parse_utils import print_table

@click.group(name="datasets")
def datasets():
    pass


@click.command(name="ls")
def ls_datasets():
    click.secho("\n[+] Collecting all datasets\n", fg='blue', bold=True)
    
    columns, values = Datasets.get_datasets(fmt="table")
    print_table(columns, values)


@click.command(name="inspect")
@click.option('--id', required=True, help='Dataset ID')
def dataset_inspect(id):
    click.secho("\n[+] Collecting information from the dataset\n", fg='blue', bold=True)
    
    dataset = Datasets.get_dataset(id)
    click.echo(dataset.get_dataset_info())
    
    
@click.command(name="columns")
@click.option('--id', required=True, help='Dataset ID')
def dataset_columns(id):
    click.secho("\n[+] Collecting columns from the dataset\n", fg='blue', bold=True)
    
    dataset = Datasets.get_dataset(id)
    click.echo(dataset.get_column_names())
    
    
@click.command(name="download")
@click.option('--id', required=True, help='Dataset ID')
@click.option('--path', required=True, help='Path where the downloaded dataset will be saved')
def dataset_download(id, path):
    click.secho("\n[+] Downloading dataset\n", fg='blue', bold=True)
    dataset = Datasets.get_dataset(id)
    dataset.download(path)


@click.command(name="upload")
@click.option('--name', required=True, help='Path where the dataset to upload is located')
@click.option('--dir_path', required=True, help='Directory where the dataset to upload is located')
@click.option('--type', required=True, help='Dataset type. It should be hf, csv or json')
def dataset_upload(name, dir_path, type):
    type = type.strip()
    assert type in ["hf", "csv", "json"], "Dataset type should be hf, csv or json"
    dir_path = Path(dir_path)
    assert dir_path.exists(), "The directory path does not exist"
    
    dataset = Dataset(
        name=name,
        directory_path=dir_path,
        dataset_type=type
    )
    
    click.secho("\n[+] Uploading dataset...\n", fg='blue', bold=True)  
    dataset.upload()
    click.secho("\n[+] Dataset uploaded\n", fg='green', bold=True)


datasets.add_command(ls_datasets)
datasets.add_command(dataset_inspect)
datasets.add_command(dataset_download)
datasets.add_command(dataset_upload)
datasets.add_command(dataset_columns)


import requests
from typing import Union, List
import numpy as np
from stochasticx.local.download_models import download_model_from_s3
import tempfile
from pathlib import Path
import uuid

from diffusers import StableDiffusionPipeline
from typing import Union, List
import time
import sys
    

def inference(
    url: str,
    headers: dict,
    prompt: Union[str, List[str]],
    img_height: int = 512,
    img_width: int = 512,
    num_inference_steps: int = 50,
    guidance_scale: float = 7.5,
    num_images_per_prompt: int = 1,
    seed: int = None
):
    response = requests.post(
        url=url,
        headers=headers,
        json={
            "prompt": prompt,
            "img_height": img_height,
            "img_width": img_width,
            "num_inference_steps": num_inference_steps,
            "guidance_scale": guidance_scale,
            "num_images_per_prompt": num_images_per_prompt,
            "seed": seed
        }
    )
    
    response.raise_for_status()
    data = response.json()
    
    images = np.array(data.get("images"))
    time = data.get("generation_time_in_secs")
    
    return images, time
    
    
def init():
    error = False
    try:
        import torch
    except:
        print("[+] Install PyTorch following the official instructions: https://pytorch.org/get-started/locally/")
        error = True
        
    if error:
        sys.exit()
    
    temp_dir = Path(tempfile.gettempdir()) / str(uuid.uuid4())
    
    print("[+] Downloading the model")
    download_model_from_s3(
        "https://stochasticai.s3.amazonaws.com/stable-diffusion/pytorch_model.zip".format(type),
        str(temp_dir)
    )
    print("[+] Model downloaded")
    
    pytorch_model_path = temp_dir / "pytorch_model"
    
    print("[+] Loading model")
    pipe = StableDiffusionPipeline.from_pretrained(
        str(pytorch_model_path), 
        revision="fp16", 
        torch_dtype=torch.float16,
        use_auth_token=True
    )
    print("[+] Model loaded")
    
    if torch.cuda.is_available():
        pipe = pipe.to("cuda")
        
    return pipe


def generate_images(
    model: StableDiffusionPipeline,
    prompt: Union[str, List[str]],
    img_height: int = 512,
    img_width: int = 512,
    num_inference_steps: int = 50,
    guidance_scale: float = 7.5,
    num_images_per_prompt: int = 1,
    seed: int = None,
    return_time=False
):
    """Do inference
    
    :param model: the Stable Diffusion pipeline
    :param prompt: the prompt
    :param img_height: height of the generated image, defaults to 512
    :param img_width: width of the generated image, defaults to 512
    :param num_inference_steps: the number of denoising steps. More denoising steps usually lead to a higher quality image at the expense of slower inference, defaults to 50
    :param guidance_scale: guidance scale, defaults to 7.5
    :param num_images_per_prompt: the number of images to generate per prompt, defaults to 1
    :param seed: Seed to make generation deterministic, defaults to None
    :param return_time: specify if time taken to generate the images should be returned, defaults to False
    :return: the output images and the time (if return time is True)
    """
    
    error = False
    try:
        import torch
    except:
        print("[+] Install PyTorch following the official instructions: https://pytorch.org/get-started/locally/")
        error = True
        
    if error:
        sys.exit()
    
    generator = None
    if seed is not None:
        generator = torch.Generator(device='cuda')
        generator = generator.manual_seed(seed)

    start_time = time.time()
    with torch.autocast("cuda"):
        output = model(
            prompt=prompt,
            height=img_height,
            width=img_width,
            num_inference_steps=num_inference_steps,
            guidance_scale=guidance_scale,
            num_images_per_prompt=num_images_per_prompt,
            generator=generator
        )
    end_time = time.time()
    
    if return_time:
        return output.images, end_time - start_time
    
    return output.images


from pathlib import Path
from typing_extensions import Required

import click
from stochasticx.models.models import Models, Model
from stochasticx.utils.parse_utils import print_table


@click.group(name="models")
def models():
    pass


@click.command(name="ls")
@click.option('--optimized', is_flag=True, help='Show only optimized models')
@click.option('--all', is_flag=True, help='Show all models')
def ls_models(optimized, all):
    if not optimized or all:
        if all:
            click.secho("\n[+] Collecting all models\n", fg='blue', bold=True)
        else:
            click.secho("\n[+] Collecting uploaded models\n", fg='blue', bold=True)
            
        columns, values = Models.get_models(fmt="table")
        print_table(columns, values)
    
    if optimized or all:
        if all:
            click.secho("\n[+] Collecting all models\n", fg='blue', bold=True)
        else:
            click.secho("\n[+] Collecting optimized models\n", fg='blue', bold=True)
            
        columns, values = Models.get_optimized_models(fmt="table")
        print_table(columns, values)


@click.command(name="inspect")
@click.option('--id', required=True, help='Model ID')
def model_inspect(id):
    click.secho("\n[+] Collecting information from this model\n", fg='blue', bold=True)
    
    model = Models.get_model(id)
    click.echo(model.get_model_info())
    
    
@click.command(name="download")
@click.option('--id', required=True, help='Model ID that you want to download')
@click.option('--path', required=True, help='Path where the downloaded model will be saved')
def model_download(id, path):
    click.secho("\n[+] Downloading the model\n", fg='blue', bold=True)
    model = Models.get_model(id)
    model.download(path)
    click.secho("\n[+] Model downloaded\n", fg='green', bold=True)


@click.command(name="upload")
@click.option('--name', required=True, help='Path where the model to upload is located')
@click.option('--dir_path', required=True, help='Directory where the model to upload is located')
@click.option('--type', required=True, help='Model type. It should be hf, pt or custom')
def model_upload(name, dir_path, type):
    type = type.strip()
    assert type in ["hf", "pt", "custom"], "Model type should be hf, pt or custom"
    dir_path = Path(dir_path)
    assert dir_path.exists(), "The directory path does not exist"
    
    model = Model(
        name=name,
        directory_path=dir_path,
        model_type=type
    )
    
    click.secho("\n[+] Uploading model...\n", fg='blue', bold=True)  
    model.upload()
    click.secho("\n[+] Model uploaded\n", fg='green', bold=True)


models.add_command(ls_models)
models.add_command(model_inspect)
models.add_command(model_download)
models.add_command(model_upload)


import subprocess
import psutil

def is_nvidia_gpu_available():
   try:
      subprocess.check_output('nvidia-smi')
      return True
   except Exception:
      return False

def get_cpu_info():
    cpu_cores= psutil.cpu_count()
    mem = psutil.virtual_memory() #(total,available,percent,used,free,...)
    
    return {"CPU_cores":cpu_cores,
            "Memory": mem[1]}

def get_gpu_info():
    gpu_name_map = {"Tesla T4":"T4"}
    gpu_info = []
    try:
        proc = subprocess.Popen(
            ["nvidia-smi",
             "--query-gpu=gpu_name,index,uuid,memory.total,memory.free,memory.used,count,utilization.gpu,utilization.memory",
              "--format=csv"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = proc.communicate()
        stdout = stdout.decode("utf-8")
        stdout = stdout.split("\n")
        stdout.pop(0)
        for l in stdout:
            tokens = l.split(", ")
            if len(tokens) > 6:
                gpu_name = tokens[0]
                if gpu_name in gpu_name_map:
                    gpu_name = gpu_name_map[tokens[0]]
                gpu_info.append({'name': gpu_name.lower(),
                                 'id': tokens[1], 
                                 'mem': tokens[3],
                                 'cores': tokens[6],
                                 'mem_free': tokens[4],
                                 'mem_used': tokens[5],
                                 'util_gpu': tokens[7],
                                 'util_mem': tokens[8]})
        return gpu_info
    except Exception:
        return None


from unicodedata import name
import docker
from typing import Dict
from stochasticx.utils.logging import configure_logger


logger = configure_logger(__name__)


def exists_container(
    container_name: str
) -> bool:
    """Check if container already exists

    :param container_name: container name
    :return: if the container exists or not
    """
    
    client = docker.from_env()
    all_containers = client.containers.list(all=True)
    
    for container in all_containers:
        if container_name == container.name:
            return True
        
    return False


def start_container(
    docker_image: str,
    ports: Dict[str, str],
    container_name: str,
    detach: bool = True,
    gpu: bool = False
):
    """Starts a Docker container locally

    :param docker_image: the Docker image
    :param ports: a dictionary speciying the ports
    :param detach: detach it or not, defaults to True
    :param gpu: use GPUs or not, defaults to False
    """
    client = docker.from_env()
    
    if exists_container(container_name):
        logger.warning("Container {} already running. Stopping and removing it...")
        stop_and_remove_container(container_name)
    
    device_requests = []
    if gpu:
        device_requests = [
            docker.types.DeviceRequest(device_ids=["all"], capabilities=[['gpu']])
        ]

    try:
        container = client.containers.run(
            docker_image, 
            detach=detach,
            ports=ports,
            device_requests=device_requests,
            name=container_name
        )
    except docker.errors.DockerException as ex:
        logger.error(str(ex))
    
    return container.id


def stop_and_remove_container(container_name):
    """Stops and removes the container

    :param container_name: container name
    """
    try:
        client = docker.from_env()
        container = client.containers.get(container_name)
        container.stop(timeout=50)
        container.remove()
    except docker.errors.DockerException as ex:
        logger.error(str(ex))
        
        
def get_logs_container(container_name):
    """Get the logs of a Docker container

    :param container_name: container name
    :return: the logs
    """
    client = docker.from_env()
    
    if exists_container(container_name):
        return client.containers.get(container_name).logs().decode("utf-8")
    else:
        return "No logs\n"
    
def get_open_ports_container(container_name):
    """Get open ports of a Docker container

    :param container_name: container name
    :return: container ports
    """
    client = docker.from_env()
    postprocess_ports = {}
    
    if exists_container(container_name):
        ports = client.containers.get(container_name).ports
        
        for container_port, host_port in ports.items():
            postprocess_ports[container_port] = host_port[0].get("HostPort")
            
        return postprocess_ports
    else:
        return {}


import click
from stochasticx.client.models import models
from stochasticx.client.datasets import datasets
from stochasticx.client.jobs import jobs
from stochasticx.client.auth import login, me
from stochasticx.client.deployments import instances, deployments
from stochasticx.client.inferences import inference
from stochasticx.client.local import stable_diffusion


@click.group()
def cli():
    pass
    
 
cli.add_command(models)
cli.add_command(datasets)
cli.add_command(jobs)
cli.add_command(login)
cli.add_command(me)
cli.add_command(instances)
cli.add_command(deployments)
cli.add_command(inference)
cli.add_command(stable_diffusion)